/*
 * Copyright 2016, Google Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jf.util.jcommander;

import com.beust.jcommander.JCommander;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import org.jf.util.ConsoleUtil;

import javax.annotation.Nonnull;
import java.util.List;

public abstract class Command {

    @Nonnull
    protected final List<JCommander> commandAncestors;

    public Command(@Nonnull List<JCommander> commandAncestors) {
        this.commandAncestors = commandAncestors;
    }

    public void usage() {
        System.out.println(new HelpFormatter()
                .width(ConsoleUtil.getConsoleWidth())
                .format(getCommandHierarchy()));
    }

    protected void setupCommand(JCommander jc) {
    }

    protected JCommander getJCommander() {
        JCommander parentJc = Iterables.getLast(commandAncestors);
        return parentJc.getCommands().get(this.getClass().getAnnotation(ExtendedParameters.class).commandName());
    }

    public List<JCommander> getCommandHierarchy() {
        List<JCommander> commandHierarchy = Lists.newArrayList(commandAncestors);
        commandHierarchy.add(getJCommander());
        return commandHierarchy;
    }

    public abstract void run();
}
